/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CAENPsdController.h"
#include "drivers/global/DriversCommands.h"
#include "controllers/common/family/Families.h"
#include <boost/thread/thread.hpp>

using namespace std;
using namespace boost;

namespace npp {

const string CAENPsdController::TYPE = "psd_controller";

/*
 * Constructor
 */
CAENPsdController::CAENPsdController(const string& name) : CAENAdcController(name) {

	setFamily(family::ACQUISITION, family::SETTING);

	// Register properties
	triggerHoldOff.init(this, NOSAVE, "trigger_holdoff");
	triggerThreshold.init(this, NOSAVE, "trigger_threshold");
	selfTrigger.init(this, NOSAVE, "self_trigger");
	chargeSensitivity.init(this, NOSAVE, "charge_sensitivity");
	shortGate.init(this, NOSAVE, "short_gate");
	longGate.init(this, NOSAVE, "long_gate");
	preGate.init(this, NOSAVE, "pre_gate");
	triggerValidationWindow.init(this, NOSAVE, "trigger_validation");
	baselineMean.init(this, NOSAVE, "base_mean");

	m_dgtzDriver.init(this, "psd_driver");

	registerFunction(TYPE);
}

/*
 * Destructor
 */
CAENPsdController::~CAENPsdController() {

}

/*
 * postConfiguration
 */
void CAENPsdController::postConfiguration() {

	registerRefresher(recordLength, &CAENPsdController::refreshRecordLengthProperty, this);
	registerRefresher(nbChannels, &CAENPsdController::refreshNbChannelsProperty, this);
	registerRefresher(dppAcqMode, &CAENPsdController::refreshDppAcqModeProperty, this);
	registerRefresher(displayChannel, &CAENPsdController::refreshDisplayChannelProperty, this);
	registerRefresher(triggerHoldOff, &CAENPsdController::refreshTriggerHoldOffProperty, this);
	registerRefresher(triggerThreshold, &CAENPsdController::refreshTriggerThresholdProperty, this);
	registerRefresher(selfTrigger, &CAENPsdController::refreshSelfTriggerProperty, this);
	registerRefresher(chargeSensitivity, &CAENPsdController::refreshChargeSensitivityProperty, this);
	registerRefresher(shortGate, &CAENPsdController::refreshShortGateProperty, this);
	registerRefresher(longGate, &CAENPsdController::refreshLongGateProperty, this);
	registerRefresher(preGate, &CAENPsdController::refreshPreGateProperty, this);
	registerRefresher(baselineMean, &CAENPsdController::refreshBaselineMean, this);
	registerRefresher(channelActive, &CAENPsdController::refreshChannelActiveProperty, this);
	registerRefresher(dcOffset, &CAENPsdController::refreshDcOffsetProperty, this);
	registerRefresher(preTrigger, &CAENPsdController::refreshPreTriggerProperty, this);
	registerRefresher(pulsePolarity, &CAENPsdController::refreshPulsePolarityProperty, this);

	updateBoardProperties();
}

void CAENPsdController::refreshRecordLengthProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->recordLength = aValue;
}

void CAENPsdController::refreshNbChannelsProperty(int32 aValue)  throw (CannotSetValue) {

	minRange = 0;
	maxRange = aValue - 1;
	m_dgtzDriver->nbChannels = aValue;
	if (displayChannel() >= aValue)
		displayChannel = aValue - 1;
}

void CAENPsdController::refreshDppAcqModeProperty(int32 aValue)  throw (CannotSetValue) {

	m_dgtzDriver->dppAcqMode = aValue;
	boost::thread td(boost::bind(&CAENPsdController::initBoard, this));
}

void CAENPsdController::refreshDisplayChannelProperty(int32 aChannel) throw (CannotSetValue) {

	float64 holdOff = m_dgtzDriver->triggerHoldOff() / 1000.0;
	if (holdOff == 0)
		holdOff = 0.2;
	triggerHoldOff.update(holdOff);

	int32 trigTrsd = m_dgtzDriver->triggerThreshold.get(aChannel);
	if (trigTrsd == 0)
		trigTrsd = 20;
	triggerThreshold.update(trigTrsd);

	selfTrigger.update(m_dgtzDriver->selfTrigger.get(aChannel));

	int32 chargeSens = m_dgtzDriver->chargeSensitivity.get(aChannel);
	chargeSensitivity.update(chargeSens);

	float64 sGate = m_dgtzDriver->shortGate.get(aChannel) / 1000.0;
	if (sGate == 0.0)
		sGate = 0.2;
	shortGate.update(sGate);

	float64 lGate = m_dgtzDriver->longGate.get(aChannel) / 1000.0;
	if (lGate == 0.0)
		lGate = 1.0;
	longGate.update(lGate);

	float64 pGate = m_dgtzDriver->preGate.get(aChannel) / 1000.0;
	if (pGate == 0)
		pGate = 0.1;
	preGate.update(pGate);

	int32 baseAvrg = m_dgtzDriver->baselineMean.get(aChannel);
	if (baseAvrg == 0)
		baseAvrg = 4;
	baselineMean.update(baseAvrg);

	channelActive.update(m_dgtzDriver->channelActive.get(aChannel));

	int32 offSet = m_dgtzDriver->dcOffset.get(aChannel);
	if (offSet == 0)
		offSet = 10;
	dcOffset.update(offSet);

	int32 preTrig = m_dgtzDriver->preTrigger.get(aChannel);
	if (preTrig == 0)
		preTrig = 500;
	preTrigger.update(preTrig);

	pulsePolarity.update(m_dgtzDriver->pulsePolarity.get(aChannel));

	initChannel(aChannel);
}

void CAENPsdController::refreshTriggerHoldOffProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerHoldOff.update((int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshTriggerThresholdProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerThreshold.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshSelfTriggerProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->selfTrigger.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshChargeSensitivityProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->chargeSensitivity.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshShortGateProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->shortGate.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshLongGateProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->longGate.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshPreGateProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->preGate.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshBaselineMean(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->baselineMean.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshChannelActiveProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->channelActive.set(displayChannel(), aValue);

	// Update channel mask
	int32 mask = 0;
	for (int32 i=0; i<m_dgtzDriver->nbChannels(); ++i) {
		if (m_dgtzDriver->channelActive.get(i))
				mask += (m_dgtzDriver->channelActive.get(i) << i);
	}
	m_dgtzDriver->channelMask = mask;
}

void CAENPsdController::refreshDcOffsetProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->dcOffset.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshPreTriggerProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->preTrigger.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::refreshPulsePolarityProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->pulsePolarity.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::updateBoardProperties() {

	modelName.update(m_dgtzDriver->modelName());
	channels.update(m_dgtzDriver->channels());
	nbChannels.update(m_dgtzDriver->nbChannels());
	minRange = 0;
	maxRange = m_dgtzDriver->nbChannels() - 1;
	dppAcqMode.update(m_dgtzDriver->dppAcqMode());
}

void CAENPsdController::writeParam() {

	m_dgtzDriver.execute(caen_psd::CAENPsdDriver::WRITE_PARAMETERS_COMMAND, false);
}

void CAENPsdController::initBoard() {

	m_dgtzDriver.execute(driver::INIT_COMMAND, false);
}

void CAENPsdController::initChannel(int32 aChannel) {

	m_dgtzDriver->triggerHoldOff.update((int32)(triggerHoldOff() * 1000));
	m_dgtzDriver->triggerThreshold.set(aChannel, triggerThreshold());
	m_dgtzDriver->selfTrigger.set(aChannel,selfTrigger());
	m_dgtzDriver->chargeSensitivity.set(aChannel, chargeSensitivity());
	m_dgtzDriver->shortGate.set(aChannel, (int32)(shortGate() * 1000));
	m_dgtzDriver->longGate.set(aChannel, (int32)(longGate() * 1000));
	m_dgtzDriver->preGate.set(aChannel, (int32)(preGate() * 1000));
	m_dgtzDriver->triggerValidationWindow.set(aChannel, triggerValidationWindow());
	m_dgtzDriver->baselineMean.set(aChannel, baselineMean());
	m_dgtzDriver->channelActive.set(aChannel, channelActive());
	m_dgtzDriver->dcOffset.set(aChannel, dcOffset());
	m_dgtzDriver->preTrigger.set(aChannel, preTrigger());
	m_dgtzDriver->pulsePolarity.set(aChannel, pulsePolarity());

	// Update channel mask
	int32 mask = 0;
	for (int32 i=0; i<m_dgtzDriver->nbChannels(); ++i) {
		if (m_dgtzDriver->channelActive.get(i))
				mask += (m_dgtzDriver->channelActive.get(i) << i);
	}
	m_dgtzDriver->channelMask = mask;

	m_dgtzDriver->dcOffset.set(aChannel, dcOffset());
	m_dgtzDriver->preTrigger.set(aChannel, preTrigger());
	m_dgtzDriver->pulsePolarity.set(aChannel, pulsePolarity());

	boost::thread td(boost::bind(&CAENPsdController::writeParam, this));
}

void CAENPsdController::clearData() {

	commandProgression = 0;
	m_dgtzDriver.execute(caen_psd::CAENPsdDriver::CLEAR_COMMAND, true);
	commandProgression = 100;
}

}
