/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RemoteDPPCoincidence.h"
#include <common/base/ServerProperties.h>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <cstdlib>

namespace npp {

using namespace std;
using namespace common;
using namespace boost;
using namespace nappli;

const std::string RemoteDPPCoincidence::TYPE = "remote_dpp_coincidence";

RemoteDPPCoincidence::RemoteDPPCoincidence(const std::string& name) :
	DPPLiveProcess(name) {

	serverEndpoint.init(this, SAVE, "nappli_server");
	dT.init(this, SAVE, "dT");
	cleanDT.init(this, SAVE, "clean_dT");
	blockSize.init(this, SAVE, "process_block_size");
	test.init(this, SAVE, "test");
	testFileName.init(this, SAVE, "test_file_name");
	xData.init(this, NOSAVE, "x_data");
	coincidenceResolution.init(this, SAVE, "coincidence_resolution");
	nOrder.init(this, NOSAVE, "n_order");
	refreshTimeS.init(this, SAVE, "refresh_time_s");
	one.init(this, NOSAVE, "one");
	order.init(this, NOSAVE, "order");
	globalRate.init(this, NOSAVE, "global_rate");
	globalCleanRate.init(this, NOSAVE, "global_clean_rate");
	detectorRate.init(this, NOSAVE, "detector_rate");
	detectorCleanRate.init(this, NOSAVE, "detector_clean_rate");
	ADCId.init(this, NOSAVE, "adc_id");
	ADCRate.init(this, NOSAVE, "adc_rate");
}

RemoteDPPCoincidence::RemoteDPPCoincidence(const RemoteDPPCoincidence& controller) :
	DPPLiveProcess(controller) {
}

RemoteDPPCoincidence::~RemoteDPPCoincidence() {
}

void RemoteDPPCoincidence::postConfiguration() {

	string endpoint = serverEndpoint();
	if (endpoint == "") {
		m_server = Application::getLocalServer();

	} else {
		m_server.reset(new Server(endpoint));
	}

	// order
	nOrder.set(7);

	order.resize(7);
	globalRate.resize(7);
	globalCleanRate.resize(7);
	detectorRate.resize(7);
	detectorCleanRate.resize(7);

	// properties
	for (int32 i = 0; i < 7; i++) {
		order.set(i, i + 2);
	}

	ADCId.resize(MAX_ADC);
	ADCRate.resize(MAX_ADC);

	for (int32 i = 0; i < MAX_ADC; i++) {
		ADCId.set(i, i);
	}

	resetData();
}

void RemoteDPPCoincidence::setDetectorControllers(const std::vector<ControllerPtr<acquisition::DetectorController> > & detectorControllers) {

	DPPLiveProcess::setDetectorControllers(detectorControllers);

	// register the list mode data
	int32 index = 0;
	for (vector<ControllerPtr<acquisition::DetectorController> >::const_iterator c = detectorControllers.begin(); c != detectorControllers.end(); ++c) {
		registerUpdater((*c)->listModeData, &RemoteDPPCoincidence::updateListModeData, this, index);
		index++;
	}
}

void RemoteDPPCoincidence::start() {

	auto_ptr<DPPReader> reader = initTest();

	resetData();

	// napplicoincidence <numor> <number of ADCs> <number of channels> <coincidence dT> <clean dT> <crateboard.txt> <detectoradc.txt> [processBlockSize] [result period]
	vector<string> args;

	// numor
	string numor;
	size_t size = fileName().size();

	if (size < 6) {
		string prefix(6 - size, '0');
		numor = prefix + fileName();
	} else {
		numor = fileName();
	}
	args.push_back(numor);
	args.push_back(lexical_cast<string>(numberOfADCs()));
	args.push_back(lexical_cast<string>(numberOfChannels()));
	args.push_back(lexical_cast<string>(dT() / 10));
	args.push_back(lexical_cast<string>(cleanDT() / 10));
	args.push_back(ServerProperties::getInstance()->getHomePath() + "/.nomadserver/crateboardcoincidence.txt");
	args.push_back(ServerProperties::getInstance()->getHomePath() + "/.nomadserver/detectoradc.txt");
	args.push_back(lexical_cast<string>(blockSize()));
	args.push_back(lexical_cast<string>(refreshTimeS()));

	// kill the application if one remains
	m_server->killAllAndWaitFor("ncoincidence");

	// start the application
	m_coincidenceApplication = m_server->start("ncoincidence", args);

	// test if application started
	if (!m_coincidenceApplication->exists()) {
		log(Level::s_Error) << "cannot start coincidence application" << endlog;
		return;
	}

	// create the publisher after the start of the application, because it is a blocking call
	m_publisher = Application::publish("dpp_publisher", 1);

	// create the subscriber after the publisher
	auto_ptr<nappli::Application::Subscriber> subscriber = m_coincidenceApplication->subscribe("dpp_coincidence_results");

	// start the test thread if necessary
	auto_ptr<boost::thread> thread = startTest(reader.get());

	// listen the result stream
	vector<double> ADCRates;
	while (subscriber->receive(ADCRates)) {
		// copying the rates
		for (int32 i = 0; i < numberOfADCs(); i++) {
			ADCRate.set(i, ADCRates[i]);
		}

		vector<double> coincidenceRates;
		if (!subscriber->receive(coincidenceRates)) {
			break;
		}

		if (coincidenceRates.size() <= 4 * 7) {
			// copying values
			for (int32 i = 0; i < 7; ++i) {
				globalRate.set(i, coincidenceRates[i]);
				globalCleanRate.set(i, coincidenceRates[7 + i]);
				detectorRate.set(i, coincidenceRates[7 * 2 + i]);
				detectorCleanRate.set(i, coincidenceRates[7 * 3 + i]);
			}

		} else {
			cerr << "problem with coincidence rates size" << endl;
		}
	}

	// wait for the termination of the application
	Application::State state = m_coincidenceApplication->waitFor();

	cout << "coincidence application terminated with state " << state << endl;

	// join the test thread
	finishTest(thread);

	// terminate the publisher
	m_publisher->terminate();
}

void RemoteDPPCoincidence::stop() {

	if (m_coincidenceApplication.get() != 0) {
		cout << *m_publisher << " end of stream" << endl;
		m_publisher->setEndOfStream();
		// we do not stop the application because the termination of the publisher will do it.
	}
}

void RemoteDPPCoincidence::raz() {
	resetData();
}

void RemoteDPPCoincidence::updateListModeData(int index) {

	if (m_publisher.get() != 0 && !m_publisher->isEndOfStream()) {
		int32 size = detectorControllers[index]->listModeData.getSize();
		int32 * memoryBlock = detectorControllers[index]->listModeData();

		// put the array into a string
		string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32) * size);

		// send the data
		m_publisher->send(data);

		cout << "updateListModeData " << index << " size " << size << endl;
	}
}

auto_ptr<DPPReader> RemoteDPPCoincidence::initTest() {

	if (test()) {
		auto_ptr<DPPReader> reader(new DPPReader(blockSize()));

		reader->openFile(testFileName());

		cout << "#ADCs " << reader->numberOfADCs() << endl;
		cout << "#channels " << reader->numberOfChannels() << endl;

		// override real values
		numberOfADCs = reader->numberOfADCs();
		numberOfChannels = reader->numberOfChannels();

		return reader;
	}

	return auto_ptr<DPPReader>();
}

void RemoteDPPCoincidence::testLoop(DPPReader * reader) {

	// loop over events
	while (!m_publisher->isEndOfStream()) {
		// reading next block
		int32_t blockSize = reader->readBlock();

		// exiting if no data
		if (blockSize == 0) {
			break;
		}

		// sending block
		int32_t * memoryBlock = reader->memoryBlock();

		// put the array into a string
		string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32_t) * blockSize);

		// send the data
		m_publisher->send(data);

		//sleep(refreshTimeS());
		usleep(10000);
	}

	cout << "finished reading test file " << testFileName() << endl;
}

std::auto_ptr<boost::thread> RemoteDPPCoincidence::startTest(DPPReader * reader) {

	return auto_ptr<thread>(new thread(bind(&RemoteDPPCoincidence::testLoop, this, reader)));
}

void RemoteDPPCoincidence::finishTest(std::auto_ptr<boost::thread> thread) {

	// join the thread
	thread->join();
}

void RemoteDPPCoincidence::resetData() {

	// resetting values
	for (int32 i = 0; i < 7; ++i) {
		globalRate.set(i, 0.0);
		globalCleanRate.set(i, 0.0);
		detectorRate.set(i, 0.0);
		detectorCleanRate.set(i, 0.0);
	}

	// setting dynamic properties
	for (int32 i = 0; i < MAX_ADC; i++) {
		ADCRate.set(i, 0.0);
	}
}

//void RemoteDPPCoincidence::executeRefreshProcess() {
//
//	while (_liveRunning) {
//		sleep(refreshTimeS());
//
//		// copying values
//		for (int32 i = 0; i < 7; ++i) {
//			globalRate.set(i, _data[i]);
//			globalCleanRate.set(i, _data[7 + i]);
//			detectorRate.set(i, _data[7 * 2 + i]);
//			detectorCleanRate.set(i, _data[7 * 3 + i]);
//		}
//
//		coincidenceData.sendEvent();
//
//		// setting dynamic properties
//		for (int32  i = 0; i < numberOfADCs(); i++) {
//			ADCRate.set(i, _ADCRates[i]);
//		}
//	}
//
//	float64 crystalRate2 = globalRate(0);
//	float64 crystalCleanRate2 = globalCleanRate(0);
//	float64 detectorRate2 = detectorRate(0);
//	float64 detectorCleanRate2 = detectorCleanRate(0);
//
//	float64 crystalRate3 = globalRate(1);
//	float64 crystalCleanRate3 = globalCleanRate(1);
//	float64 detectorRate3 = detectorRate(1);
//	float64 detectorCleanRate3 = detectorCleanRate(1);
//
//	ostringstream ss1, ss2, ss3;
//
//	ss1 << "  Order"
//			<< "        Crystal (c/s)"
//			<< "  Crystal clean (c/s)"
//			<< "       Detector (c/s)"
//			<< " Detector clean (c/s)";
//
//	ss2 << "      2 " << boost::format("%20.3d") % crystalRate2 << " " << boost::format("%20.3d") % crystalCleanRate2 << " " << boost::format("%20.3d") % detectorRate2 << " " << boost::format("%20.3d") % detectorCleanRate2;
//	ss3 << "      3 " << boost::format("%20.3d") % crystalRate3 << " " << boost::format("%20.3d") % crystalCleanRate3 << " " << boost::format("%20.3d") % detectorRate3 << " " << boost::format("%20.3d") % detectorCleanRate3;
//
//	log(Level::s_Info) << ss1.str() << endlog;
//	log(Level::s_Info) << ss2.str() << endlog;
//	log(Level::s_Info) << ss3.str() << endlog;
//}

}
